# Copyright (c) OpenMMLab. All rights reserved.
from .dgcnn_fa_module import DGCNNFAModule
from .dgcnn_fp_module import DGCNNFPModule
from .dgcnn_gf_module import DGCNNGFModule

__all__ = ['DGCNNFAModule', 'DGCNNFPModule', 'DGCNNGFModule']
